import datetime
import email.message
import math
import logging
import unittest.mock as mock
from typing import Callable, Any

import pytest
from pytest import LogCaptureFixture
from pysimplesoap.simplexml import SimpleXMLElement

import debianbts as bts
from debianbts import Bugreport

logger = logging.getLogger(__name__)


@pytest.fixture
def create_bugreport() -> Callable[..., Bugreport]:
    def factory(**kwargs: Any) -> Bugreport:
        bugreport = bts.Bugreport()
        for k, v in kwargs.items():
            setattr(bugreport, k, v)
        return bugreport

    return factory


def test_get_usertag_empty() -> None:
    """get_usertag should return empty dict if no bugs are found."""
    d = bts.get_usertag("thisisatest@debian.org")
    assert d == dict()


def test_get_usertag() -> None:
    """get_usertag should return dict with tag(s) and buglist(s)."""
    d = bts.get_usertag("debian-python@lists.debian.org")
    assert isinstance(d, dict)
    for k, v in d.items():
        assert isinstance(k, str)
        assert isinstance(v, list)
        for bug in v:
            assert isinstance(bug, int)


def test_get_usertag_args(caplog: LogCaptureFixture) -> None:
    # no tags
    tags = bts.get_usertag("debian-python@lists.debian.org")
    assert len(tags) > 2

    randomKey0 = list(tags.keys())[0]
    randomKey1 = list(tags.keys())[1]

    # one tags
    tags = bts.get_usertag("debian-python@lists.debian.org", [randomKey0])
    assert len(tags) == 1

    # two tags
    tags = bts.get_usertag(
        "debian-python@lists.debian.org", [randomKey0, randomKey1]
    )
    assert len(tags) == 2


def test_get_bugs_empty(caplog: LogCaptureFixture) -> None:
    """get_bugs should return empty list if no matching bugs where found."""
    bugs = bts.get_bugs(package="thisisatest")
    assert bugs == []


def test_get_bugs(caplog: LogCaptureFixture) -> None:
    """get_bugs should return list of bugnumbers."""
    bugs = bts.get_bugs(submitter="venthur@debian.org")
    assert len(bugs) != 0
    assert isinstance(bugs, list)
    for i in bugs:
        assert isinstance(i, int)


def test_newest_bugs() -> None:
    """newest_bugs should return list of bugnumbers."""
    bugs = bts.newest_bugs(10)
    assert isinstance(bugs, list)
    for i in bugs:
        assert isinstance(i, int)


def test_newest_bugs_amount() -> None:
    """newest_bugs(amount) should return a list of len 'amount'."""
    for i in 0, 1, 10:
        bugs = bts.newest_bugs(i)
        assert len(bugs) == i


def test_get_bug_log() -> None:
    """get_bug_log should return the correct data types."""
    bl = bts.get_bug_log(223344)
    assert isinstance(bl, list)
    for i in bl:
        assert isinstance(i, dict)
        assert "attachments" in i
        assert isinstance(i["attachments"], list)
        assert "body" in i
        assert isinstance(i["body"], str)
        assert "header" in i
        assert isinstance(i["header"], str)
        assert "msg_num" in i
        assert isinstance(i["msg_num"], int)


def test_get_bug_log_with_attachments() -> None:
    """get_bug_log should include attachments"""
    buglogs = bts.get_bug_log(400000)
    for bl in buglogs:
        assert "attachments" in bl


def test_bug_log_message() -> None:
    """dict returned by get_bug_log has a email.Message field"""
    buglogs = bts.get_bug_log(400012)
    for buglog in buglogs:
        assert "message" in buglog
        msg = buglog["message"]
        assert isinstance(msg, email.message.Message)
        assert "Subject" in msg
        if not msg.is_multipart():
            assert isinstance(msg.get_payload(), str)


def test_bug_log_message_unicode() -> None:
    """test parsing of bug_log mail with non ascii characters"""
    buglogs = bts.get_bug_log(773321)
    buglog = buglogs[2]
    msg = buglog["message"]
    assert isinstance(msg, email.message.Message)
    msg_payload = msg.get_payload()
    assert isinstance(msg_payload, str)
    assert "é" in msg_payload


def test_empty_get_status() -> None:
    """get_status should return empty list if bug doesn't exits"""
    bugs = bts.get_status(0)
    assert isinstance(bugs, list)
    assert len(bugs) == 0


def test_get_status_params(caplog: LogCaptureFixture) -> None:
    BUG = 223344
    BUG2 = 334455

    bugs = bts.get_status(BUG)
    assert isinstance(bugs, list)
    assert len(bugs) == 1

    bugs = bts.get_status([BUG, BUG2])
    assert isinstance(bugs, list)
    assert len(bugs) == 2

    bugs = bts.get_status((BUG, BUG2))
    assert isinstance(bugs, list)
    assert len(bugs) == 2


def test_sample_get_status() -> None:
    """test retrieving of a "known" bug status"""
    bugs = bts.get_status(486212)
    assert len(bugs) == 1
    bug = bugs[0]
    assert bug.bug_num == 486212
    assert bug.date == datetime.datetime(2008, 6, 14, 10, 30, 2)
    assert bug.subject.startswith("[reportbug-ng] segm")
    assert bug.package == "reportbug-ng"
    assert bug.severity == "normal"
    assert bug.tags == ["help"]
    assert bug.blockedby == []
    assert bug.blocks == []
    assert bug.summary == ""
    assert bug.location == "archive"
    assert bug.source == "reportbug-ng"
    assert bug.log_modified == datetime.datetime(2008, 8, 17, 7, 26, 22)
    assert bug.pending == "done"
    assert bug.done
    assert bug.done_by == "Bastian Venthur <venthur@debian.org>"
    assert bug.archived
    assert bug.found_versions == ["reportbug-ng/0.2008.06.04"]
    assert bug.fixed_versions == ["reportbug-ng/1.0"]
    assert bug.affects == []


def test_done_by_decoding() -> None:
    """Done by is properly base64 decoded when needed."""
    # no base64 encoding
    bug = bts.get_status(486212)[0]
    assert bug.done_by == "Bastian Venthur <venthur@debian.org>"

    # base64 encoding
    bug = bts.get_status(938128)[0]
    assert bug.done_by == "Ondřej Nový <onovy@debian.org>"


def test_bug_str(
    create_bugreport: Callable[..., Bugreport],
) -> None:
    """test string conversion of a Bugreport"""
    b1 = create_bugreport(package="foo-pkg", bug_num=12222)
    s = str(b1)
    assert isinstance(s, str)  # byte string in py2, unicode in py3
    assert "bug_num: 12222\n" in s
    assert "package: foo-pkg\n" in s


def test_get_status_affects() -> None:
    """test a bug with "affects" field"""
    bugs = bts.get_status([290501, 770490])
    assert len(bugs) == 2
    assert bugs[0].affects == []
    assert bugs[1].affects == ["conkeror"]


@mock.patch.object(bts.debianbts, "_build_soap_client")
def test_status_batches_large_bug_counts(
    mock_build_client: Any,
) -> None:
    """get_status should perform requests in batches to reduce server load."""
    mock_build_client.return_value = mock_client = mock.Mock()
    mock_client.call.return_value = SimpleXMLElement("<a><s-gensym3/></a>")
    nr = bts.BATCH_SIZE + 10.0
    calls = int(math.ceil(nr / bts.BATCH_SIZE))
    bts.get_status([722226] * int(nr))
    assert mock_client.call.call_count == calls


@mock.patch.object(bts.debianbts, "_build_soap_client")
def test_status_batches_multiple_arguments(
    mock_build_client: Any,
) -> None:
    """get_status should batch multiple arguments into one request."""
    mock_build_client.return_value = mock_client = mock.Mock()
    mock_client.call.return_value = SimpleXMLElement("<a><s-gensym3/></a>")
    batch_size = bts.BATCH_SIZE

    calls = 1
    bts.get_status(list(range(batch_size)))
    assert mock_client.call.call_count == calls

    calls += 2
    bts.get_status(list(range(batch_size + 1)))
    assert mock_client.call.call_count == calls


def test_comparison(create_bugreport: Callable[..., Bugreport]) -> None:
    """comparison of two bugs"""
    b1 = create_bugreport(severity="normal", archived=True)
    b2 = create_bugreport(severity="normal", archived=False, done=True)
    assert b2 > b1
    assert b2 >= b1
    assert b2 != b1
    assert not (b2 == b1)
    assert not (b2 <= b1)
    assert not (b2 < b1)


def test_comparison_equal(create_bugreport: Callable[..., Bugreport]) -> None:
    """comparison of two bug which are equal regarding their
    relative order"""
    b1 = create_bugreport(severity="normal", archived=False, done=True)
    b2 = create_bugreport(severity="normal", archived=False, done=True)
    assert not (b2 > b1)
    assert b2 >= b1
    assert b2 == b1
    assert not (b2 < b1)
    assert b2 <= b1


def test_get_bugs_int_bugs() -> None:
    """It is possible to pass a list of bug number to get_bugs"""
    bugs = bts.get_bugs(bugs=[400010, 400012], archive="1")
    assert set(bugs) == {400010, 400012}


def test_get_bugs_single_int_bug() -> None:
    """bugs parameter in get_bugs can be a list of int or a int"""
    bugs1 = bts.get_bugs(bugs=400040, archive="1")
    bugs2 = bts.get_bugs(bugs=[400040], archive="1")
    assert bugs1 == bugs2


def test_get_bugs_archived() -> None:
    """archive tristate."""
    # the parameter is rather undocumented. with trial and error i found
    # out that it takes a string with those three values. everything
    # else will be interpreted as "1"
    bugs_unarchived = bts.get_bugs(src="python-debianbgs", archive="0")
    bugs_archived = bts.get_bugs(src="python-debianbgs", archive="1")
    bugs_both = bts.get_bugs(src="python-debianbgs", archive="both")
    assert len(bugs_both) == len(bugs_unarchived) + len(bugs_archived)


def test_get_bugs_archived_default() -> None:
    """Return un-archived bugs by default."""
    bugs_unarchived = bts.get_bugs(src="python-debianbgs", archive="0")
    bugs_default = bts.get_bugs(src="python-debianbgs")
    assert len(bugs_default) == len(bugs_unarchived)


def test_mergedwith() -> None:
    """Mergedwith is always a list of int."""
    # this one is merged with two other bugs
    m = bts.get_status(486212)[0].mergedwith
    assert len(m) == 2
    for i in m:
        assert isinstance(i, int)
    # this one was merged with one bug
    m = bts.get_status(433550)[0].mergedwith
    assert len(m) == 1
    assert isinstance(m[0], int)
    # this one was not merged
    m = bts.get_status(474955)[0].mergedwith
    assert m == list()


def test_base64_status_fields() -> None:
    """fields in bug status are sometimes base64-encoded"""
    bug = bts.get_status(711111)[0]
    assert isinstance(bug.originator, str)
    assert bug.originator.endswith("gmail.com>")
    assert "ł" in bug.originator


def test_base64_buglog_body() -> None:
    """buglog body is sometimes base64 encoded"""
    buglog = bts.get_bug_log(773321)
    body1 = buglog[1]["body"]
    body2 = buglog[2]["body"]
    assert isinstance(body1, str)
    assert isinstance(body2, str)
    assert "é" in body2


def test_string_status_originator() -> None:
    """test reading of bug status originator that is not base64-encoded"""
    bug = bts.get_status(711112)[0]
    assert isinstance(bug.originator, str)
    assert bug.originator.endswith("debian.org>")


def test_unicode_conversion_in_str() -> None:
    """string representation must deal with unicode correctly."""
    [bug] = bts.get_status(773321)
    bug.__str__()


def test_regression_588954() -> None:
    """Get_bug_log must convert the body correctly to unicode."""
    bts.get_bug_log(582010)


def test_version() -> None:
    assert isinstance(bts.__version__, str)


def test_regression_590073() -> None:
    """bug.blocks is sometimes a str sometimes an int."""
    # test the int case
    # TODO: test the string case
    bts.get_status(568657)


def test_regression_590725() -> None:
    """bug.body utf sometimes contains invalid continuation bytes."""
    bts.get_bug_log(578363)
    bts.get_bug_log(570825)


def test_regression_670446() -> None:
    """affects should be split by ','"""
    bug = bts.get_status(657408)[0]
    assert bug.affects == ["epiphany-browser-dev", "libwebkit-dev"]


def test_regression_799528() -> None:
    """fields of buglog are sometimes base64 encoded."""
    # bug with base64 encoding originator
    [bug] = bts.get_status(711111)
    assert "ł" in bug.originator
    # bug with base64 encoding subject
    [bug] = bts.get_status(779005)
    assert "‘" in bug.subject


def test_regression_917165() -> None:
    bts.get_bug_log(887978)


def test_regression_917258() -> None:
    bts.get_bug_log(541147)



import threading
import logging

import debianbts as bts

logger = logging.getLogger(__name__)


class TestThreading:
    """this class tests that the module is thread safe"""

    def test_multithreading(self) -> None:
        self._thread_failed = False
        threads = [
            threading.Thread(target=self._get_bugs_thread, args=(pkg,))
            for pkg in ("python3-gst-1.0", "libsoxr0")
        ] + [
            threading.Thread(target=self._get_bug_log_thread, args=(bug_n,))
            for bug_n in (300000, 300001)
        ]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert not self._thread_failed

    def _get_bugs_thread(self, pkg: str) -> None:
        try:
            bts.get_bugs(package=pkg)
        except Exception:
            self._thread_failed = True
            logger.exception("Threaded get_bugs() call failed.")

    def _get_bug_log_thread(self, bug_num: int) -> None:
        try:
            bts.get_bug_log(bug_num)
        except Exception:
            self._thread_failed = True
            logger.exception("Threaded get_bug_log() call failed.")



# please follow the semver semantics, i.e. MAJOR.MINOR.PATCH where
# MAJOR: incompatible API changes
# MINOR: add backwards-compatible functionality
# PATCH: backwards-compatible bug fixes.
__version__ = "4.0.1"



#!/usr/bin/env python

"""
Query Debian's Bug Tracking System (BTS).

This module provides a layer between Python and Debian's BTS. It provides
methods to query the BTS using the BTS' SOAP interface, and the Bugreport class
which represents a bugreport from the BTS.
"""


from __future__ import annotations

import base64
import email.feedparser
import email.policy
from datetime import datetime
import os
import logging
from typing import Any, Iterable

from pysimplesoap.client import SoapClient
from pysimplesoap.simplexml import SimpleXMLElement


logger = logging.getLogger(__name__)


# Support running from Debian infrastructure
ca_path = "/etc/ssl/ca-debian"
if os.path.isdir(ca_path):
    os.environ["SSL_CERT_DIR"] = ca_path


# Setup the soap server
# Default values
URL = "https://bugs.debian.org/cgi-bin/soap.cgi"
NS = "Debbugs/SOAP/V1"
BTS_URL = "https://bugs.debian.org/"
# Max number of bugs to send in a single get_status request
BATCH_SIZE = 500

SEVERITIES = {
    "critical": 7,
    "grave": 6,
    "serious": 5,
    "important": 4,
    "normal": 3,
    "minor": 2,
    "wishlist": 1,
}


class Bugreport:
    """Represents a bugreport from Debian's Bug Tracking System.

    A bugreport object provides all attributes provided by the SOAP interface.
    Most of the attributes are strings, the others are marked.

    Attributes
    ----------
    bug_num : int
        The bugnumber
    severity : str
        Severity of the bugreport
    tags : list[str]
        Tags of the bugreport
    subject : str
        The subject/title of the bugreport
    originator : str
        Submitter of the bugreport
    mergedwith : list[int]
        List of bugnumbers this bug was merged with
    package : str
        Package of the bugreport
    source : str
        Source package of the bugreport
    date : datetime
        Date of bug creation
    log_modified : datetime
        Date of update of the bugreport
    done : boolean
        Is the bug fixed or not
    done_by : str | None
        Name and Email or None
    archived : bool
        Is the bug archived or not
    unarchived : bool
        Was the bug unarchived or not
    fixed_versions : list[str]
        List of versions, can be empty even if bug is fixed
    found_versions : list[str]
        List of version numbers where bug was found
    forwarded : str
        A URL or email address
    blocks: list[int]
        List of bugnumbers this bug blocks
    blockedby : list[int]
        List of bugnumbers which block this bug
    pending : str
        Either 'pending' or 'done'
    msgid : str
        Message ID of the bugreport
    owner : str
        Who took responsibility for fixing this bug
    location : str
        Either 'db-h' or 'archive'
    affects : list[str]
        List of Packagenames
    summary : str
        Arbitrary text
    """

    def __init__(self) -> None:
        self.originator: str
        self.date: datetime
        self.subject: str
        self.msgid: str
        self.package: str
        self.tags: list[str]
        self.done: bool
        self.done_by: str | None
        self.forwarded: str
        self.mergedwith: list[int]
        self.severity: str
        self.owner: str
        self.found_versions: list[str]
        self.fixed_versions: list[str]
        self.blocks: list[int]
        self.blockedby: list[int]
        self.unarchived: bool
        self.summary: str
        self.affects: list[str]
        self.log_modified: datetime
        self.location: str
        self.archived: bool
        self.bug_num: int
        self.source: str
        self.pending: str
        # The ones below are also there but not used
        # self.fixed = None
        # self.found = None
        # self.fixed_date = None
        # self.found_date = None
        # self.keywords = None
        # self.id = None

    def __str__(self) -> str:
        s = "\n".join(
            f"{key}: {value}" for key, value in self.__dict__.items()
        )
        return s + "\n"

    def __lt__(self, other: Bugreport) -> bool:
        """Compare a bugreport with another.

        The more open and urgent a bug is, the greater the bug is:

            outstanding > resolved > archived

            critical > grave > serious > important > normal > minor > wishlist.

        Openness always beats urgency, eg an archived bug is *always* smaller
        than an outstanding bug.

        This sorting is useful for displaying bugreports in a list and sorting
        them in a useful way.

        """
        return self._get_value() < other._get_value()

    def __le__(self, other: Bugreport) -> bool:
        return not self.__gt__(other)

    def __gt__(self, other: Bugreport) -> bool:
        return self._get_value() > other._get_value()

    def __ge__(self, other: Bugreport) -> bool:
        return not self.__lt__(other)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Bugreport):
            return NotImplemented
        return self._get_value() == other._get_value()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Bugreport):
            return NotImplemented
        return not self.__eq__(other)

    def _get_value(self) -> int:
        if self.archived:
            # archived and done
            val = 0
        elif self.done:
            # not archived and done
            val = 10
        else:
            # not done
            val = 20
        val += SEVERITIES[self.severity]
        return val


def get_status(
    nrs: int | list[int] | tuple[int, ...],
) -> list[Bugreport]:
    """Returns a list of Bugreport objects.

    Given a list of bug numbers this method returns a list of Bugreport
    objects.

    Parameters
    ----------
    nrs
        The bugnumbers

    Returns
    -------
    list[Bugreport]
        list of Bugreport objects

    """
    numbers: list[int]
    if not isinstance(nrs, (list, tuple)):
        numbers = [nrs]
    else:
        numbers = list(nrs)

    # Process the input in batches to avoid hitting resource limits on
    # the BTS
    soap_client = _build_soap_client()
    bugs = []
    for i in range(0, len(numbers), BATCH_SIZE):
        slice_ = numbers[i:i + BATCH_SIZE]
        # I build body by hand, pysimplesoap doesn't generate soap Arrays
        # without using wsdl
        method_el = SimpleXMLElement("<get_status></get_status>")
        _build_int_array_el("arg0", method_el, slice_)
        reply = soap_client.call("get_status", method_el)
        for bug_item_el in reply("s-gensym3").children() or []:
            bug_el = bug_item_el.children()[1]
            bugs.append(_parse_status(bug_el))
    return bugs


def get_usertag(
    email: str,
    tags: None | list[str] | tuple[str, ...] = None,
) -> dict[str, list[int]]:
    """Get buglists by usertags.

    Parameters
    ----------
    email
    tags
        If tags are given the dictionary is limited to the matching tags, if no
        tags are given all available tags are returned.

    Returns
    -------
    dict[str, list[int]]
        a mapping of usertag -> buglist

    """
    if tags is None:
        tags = []

    reply = _soap_client_call("get_usertag", email, *tags)
    map_el = reply("s-gensym3")
    mapping = {}
    # element <s-gensys3> in response can have standard type
    # xsi:type=apachens:Map (example, for email debian-python@lists.debian.org)
    # OR no type, in this case keys are the names of child elements and
    # the array is contained in the child elements
    type_attr = map_el.attributes().get("xsi:type")
    if type_attr and type_attr.value == "apachens:Map":
        for usertag_el in map_el.children() or []:
            tag = str(usertag_el("key"))
            buglist_el = usertag_el("value")
            mapping[tag] = [int(bug) for bug in buglist_el.children() or []]
    else:
        for usertag_el in map_el.children() or []:
            tag = usertag_el.get_name()
            mapping[tag] = [int(bug) for bug in usertag_el.children() or []]
    return mapping


def get_bug_log(
    nr: int,
) -> list[dict[str, str | list[Any] | int | email.message.Message]]:
    """Get Buglogs.

    A buglog is a dictionary with the following mappings:
        * "header" => string
        * "body" => string
        * "attachments" => list
        * "msg_num" => int
        * "message" => email.message.Message

    Parameters
    ----------
    nr
        the bugnumber

    Returns
    -------
    list[dict[str, str | list[Any] | int | email.message.Message]]
        list of buglogs

    """
    reply = _soap_client_call("get_bug_log", nr)
    items_el = reply("soapenc:Array")
    buglogs = []
    for buglog_el in items_el.children():
        buglog: dict[str, str | list[Any] | int | email.message.Message] = {}
        header = _parse_string_el(buglog_el("header"))
        body = _parse_string_el(buglog_el("body"))
        msg_num = int(buglog_el("msg_num"))
        # server always returns an empty attachments array ?
        attachments: list[Any] = []

        mail_parser = email.feedparser.BytesFeedParser(
            policy=email.policy.SMTP
        )
        mail_parser.feed(header.encode())
        mail_parser.feed(b"\n\n")
        mail_parser.feed(body.encode())
        message = mail_parser.close()

        buglog = {
            "header": header,
            "body": body,
            "msg_num": msg_num,
            "attachments": attachments,
            "message": message,
        }

        buglogs.append(buglog)
    return buglogs


def newest_bugs(amount: int) -> list[int]:
    """Returns the newest bugs.

    This method can be used to query the BTS for the n newest bugs.

    Parameters
    ----------
    amount
        the number of desired bugs. E.g. if `amount` is 10 the method will
        return the 10 latest bugs.

    Returns
    -------
    list[int]
        the bugnumbers

    """
    reply = _soap_client_call("newest_bugs", amount)
    items_el = reply("soapenc:Array")
    return [int(item_el) for item_el in items_el.children() or []]


def get_bugs(
    **kwargs: str | int | list[int],
) -> list[int]:
    """Get list of bugs matching certain criteria.

    The conditions are defined by the keyword arguments.

    Arguments
    ---------
    kwargs
        Possible keywords are:
            * "package": bugs for the given package
            * "submitter": bugs from the submitter
            * "maint": bugs belonging to a maintainer
            * "src": bugs belonging to a source package
            * "severity": bugs with a certain severity
            * "status": can be either "done", "forwarded", or "open"
            * "tag": see http://www.debian.org/Bugs/Developer#tags for
              available tags
            * "owner": bugs which are assigned to `owner`
            * "bugs": takes single int or list of bugnumbers, filters the list
              according to given criteria
            * "correspondent": bugs where `correspondent` has sent a mail to
            * "archive": takes a string: "0" (unarchived), "1" (archived) or
              "both" (un- and archived). if omitted, only returns un-archived
              bugs.

    Returns
    -------
    list[int]
        the bugnumbers

    Examples
    --------
        >>> get_bugs(package='gtk-qt-engine', severity='normal')
        [12345, 23456]

    """
    # flatten kwargs to list:
    # {'foo': 'bar', 'baz': 1} -> ['foo', 'bar','baz', 1]
    args = []
    for k, v in kwargs.items():
        args.extend([k, v])

    # pysimplesoap doesn't generate soap Arrays without using wsdl
    # I build body by hand, converting list to array and using standard
    # pysimplesoap marshalling for other types
    method_el = SimpleXMLElement("<get_bugs></get_bugs>")
    for arg_n, kv in enumerate(args):
        arg_name = "arg" + str(arg_n)
        if isinstance(kv, (list, tuple)):
            _build_int_array_el(arg_name, method_el, kv)
        else:
            method_el.marshall(arg_name, kv)

    soap_client = _build_soap_client()
    reply = soap_client.call("get_bugs", method_el)
    items_el = reply("soapenc:Array")
    return [int(item_el) for item_el in items_el.children() or []]


def _parse_status(bug_el: SimpleXMLElement) -> Bugreport:
    """Return a bugreport object from a given status xml element

    Parameters
    ----------
    bug_el
        a status XML element

    Returns
    -------
    Bugreport
        a Bugreport object

    """
    bug = Bugreport()

    # plain fields
    for field in (
        "originator",
        "subject",
        "msgid",
        "package",
        "severity",
        "owner",
        "summary",
        "location",
        "source",
        "pending",
        "forwarded",
    ):
        setattr(bug, field, _parse_string_el(bug_el(field)))

    bug.date = datetime.utcfromtimestamp(float(bug_el("date")))
    bug.log_modified = datetime.utcfromtimestamp(float(bug_el("log_modified")))
    bug.tags = [tag for tag in str(bug_el("tags")).split()]
    bug.done = _parse_bool(bug_el("done"))
    bug.done_by = _parse_string_el(bug_el("done")) if bug.done else None
    bug.archived = _parse_bool(bug_el("archived"))
    bug.unarchived = _parse_bool(bug_el("unarchived"))
    bug.bug_num = int(bug_el("bug_num"))
    bug.mergedwith = [int(i) for i in str(bug_el("mergedwith")).split()]
    bug.blockedby = [int(i) for i in str(bug_el("blockedby")).split()]
    bug.blocks = [int(i) for i in str(bug_el("blocks")).split()]

    bug.found_versions = [
        str(el) for el in bug_el("found_versions").children() or []
    ]
    bug.fixed_versions = [
        str(el) for el in bug_el("fixed_versions").children() or []
    ]
    affects = [_f for _f in str(bug_el("affects")).split(",") if _f]
    bug.affects = [a.strip() for a in affects]
    # Also available, but unused or broken
    # bug.keywords = [keyword for keyword in
    #                 str(bug_el('keywords')).split()]
    # bug.fixed = _parse_crappy_soap(tmp, "fixed")
    # bug.found = _parse_crappy_soap(tmp, "found")
    # bug.found_date = \
    #     [datetime.utcfromtimestamp(i) for i in tmp["found_date"]]
    # bug.fixed_date = \
    #     [datetime.utcfromtimestamp(i) for i in tmp["fixed_date"]]
    return bug


_soap_client_kwargs = {
    "location": URL,
    "action": "",
    "namespace": NS,
    "soap_ns": "soap",
}


def set_soap_proxy(proxy_arg: str) -> None:
    """Set proxy for SOAP client.

    You must use this method after import to set the proxy.

    Parameters
    ----------
    proxy_arg

    """
    _soap_client_kwargs["proxy"] = proxy_arg


def set_soap_location(url: str) -> None:
    """Set location URL for SOAP client

    You may use this method after import to override the default URL.

    Parameters
    ----------
    url
        default URL

    """
    _soap_client_kwargs["location"] = url


def get_soap_client_kwargs() -> dict[str, str]:
    """Returns SOAP client kwargs.

    Returns
    -------
    dict[str, str]
        the SOAP client kwargs

    """
    return _soap_client_kwargs


def _build_soap_client() -> SoapClient:
    """Factory method that creates a SoapClient.

    For thread-safety we create SoapClients on demand instead of using a
    module-level one.

    Returns
    -------
    SoapClient
        a SoapClient instance

    """
    return SoapClient(**_soap_client_kwargs)


def _convert_soap_method_args(*args: Iterable[Any]) -> list[tuple[str, Any]]:
    """Convert arguments to be consumed by a SoapClient method

    Parameters
    ----------
    *args
        any argument

    Returns
    -------
    list[tuple[str, Any]]
        the converted arguments

    Examples
    --------
    Soap client required a list of named arguments:

        >>> _convert_soap_method_args('a', 1)
        [('arg0', 'a'), ('arg1', 1)]

    """
    soap_args = []
    for arg_n, arg in enumerate(args):
        soap_args.append(("arg" + str(arg_n), arg))
    return soap_args


def _soap_client_call(method_name: str, *args: Any) -> Any:
    """Wrapper to call SoapClient method

    Parameters
    ----------
    method_name
        the method name
    *args

    Returns
    -------
    Any

    """
    # a new client instance is built for threading issues
    soap_client = _build_soap_client()
    soap_args = _convert_soap_method_args(*args)
    return getattr(soap_client, method_name)(*soap_args)


def _build_int_array_el(
    el_name: str,
    parent: SimpleXMLElement,
    list_: list[Any],
) -> SimpleXMLElement:
    """Build Array as child of parent.

    More specifically: Build a soapenc:Array made of ints called `el_name` as a
    child of `parent`.

    Parameters
    ----------
    el_name
    parent
    list

    Returns
    -------
    SimpleXMLElement

    """
    el = parent.add_child(el_name)
    el.add_attribute(
        "xmlns:soapenc", "http://schemas.xmlsoap.org/soap/encoding/"
    )
    el.add_attribute("xsi:type", "soapenc:Array")
    el.add_attribute("soapenc:arrayType", f"xsd:int[{len(list_):d}]")
    for item in list_:
        item_el = el.add_child("item", str(item))
        item_el.add_attribute("xsi:type", "xsd:int")
    return el


def _parse_bool(el: SimpleXMLElement) -> bool:
    """Parse a boolean value from a XML element.

    Parameters
    ----------
    el
        the element to parse

    Returns
    -------
    bool
        the parsed value

    """
    value = str(el)
    return not value.strip() in ("", "0")


def _parse_string_el(el: SimpleXMLElement) -> str:
    """Read a string element, maybe encoded in base64.

    Parameters
    ----------
    el
        the element to parse

    Returns
    -------
    str
        the parsed value

    """
    value = str(el)
    el_type = el.attributes().get("xsi:type")
    if el_type and el_type.value == "xsd:base64Binary":
        tmp = base64.b64decode(value)
        value = tmp.decode("utf-8", errors="replace")
    return value



import logging

logger = logging.getLogger("debianbts")
logging.basicConfig(
    format="%(asctime)s %(levelname)s %(name)s %(message)s",
    level=logging.WARNING,
)


def main() -> None:
    logger.warning("Not implemented yet, sorry!")


if __name__ == "__main__":
    main()
